# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
class UnknownFlatbufferIdWarning(Warning):
    pass


class BufferSizeWarning(Warning):
    pass
